/*
 * do_whois.c
 *
 * Copyright  2001 Dan Anderson <dan@drydog.com> All rights reserved.
 * Copyright  2000 Joao Cabral <jcnc@dhis.org> All rights reserved.
 * See file COPYING for licensing terms.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */


#ifndef lint
static char rcsid[] = "@(#) $Id: do_whois.c,v 1.9 2004/01/04 05:33:49 dan Exp $";
#endif

#define	SWHOISD_CONF		SYSCONFDIR "/swhoisd.conf"
#define	SWHOISD_hdr		SYSCONFDIR "/swhoisd.header"
#define	SWHOISD_FOOTER		SYSCONFDIR "/swhoisd.footer"
#define	MAX_HANDLE		254


#include "config.h"
#include "in.swhoisd.h"
#include <stdio.h>
#include <ctype.h>		/* is*() */
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>		/* sleep() */
#include <syslog.h>		/* LOG_ERR, syslog() */


/*
 * String Utilities
 */
#ifndef isblank
/* GNU isblank() is not on Solaris: */
#define	isblank(c) (((c) == ' ') || ((c) == '\t'))
#endif

/* Return non-zero (true) if NUL or return or newline character: */
#define	isendofline(c) (((c) == '\0') || ((c) == '\r') || ((c) == '\n'))


/*
 * Global Data Structures
 */

/* Saved list of handles matching query */
struct handle_list {
	char handle[MAX_HANDLE];
	struct handle_list *next;
	};


/*
 * Convert a string to lowercase
 */
static void
lowercase_string(char *s) {
	if (s == NULL) {
		print_syslog(LOG_ERR, "lowercase_string() passed NULL\n");
		return;
	}
	while (*s != '\0') {
		*s = tolower(*s);
		++s;
	}
}


/*
 * print_file_if_present()
 * Print a file to stdout if found.
 * Skip lines beginning with "#"
 * Convert "\n" to "\r\n" when printing (required by Internet standard).
 * If found, return 0, else non-zero.
 */
static int
print_file_if_present(char *filename, FILE *out_fd) {
	FILE *text_fp = NULL;
	char line[MAXLINE];

	if ((filename == NULL) || (out_fd == NULL)) {
		print_syslog(LOG_ERR, "print_file_if_present() passed NULL\n");
		return (-1); /* Failed */
	}

	/* Open file */
	text_fp = fopen(filename, "r");
	if (text_fp == NULL) {
		print_syslog(LOG_ERR, "fopen(test_fp) failed; errno %s\n",
			strerror(errno));
		return (-1); /* Failed */
	}

	/* Read and write file */
	(void) memset(line, '\0', sizeof (line));
	(void) fgets(line, sizeof (line) - 2, text_fp);
	while (!feof(text_fp) && !ferror(text_fp)) {
		/* skip comment line, if any: */
		int i = 0;
		while (isblank(line[i])) {
			++i;
		}
		if (line[i] != '#') { /* not a comment line */
			int	len = strlen(line);
			/* Convert "\n" to "\r\n\0": */
			if (line[len -1] == '\n') {
				line[len - 1] = '\r';
				line[len] = '\n';
				line[len + 1] = '\0';
			}
			(void) fputs(line, out_fd);
		}

		/* Read next line, if any: */
		(void) fgets(line, sizeof (line) - 2, text_fp);
	}

	(void) fclose(text_fp);
	return (0); /* OK */
}


/*
 * Free a link list of handles and its base pointer.
 * "Cover our tracks" by NULL-ing out pointers to guarantee they
 * will never be used again.
 */
static void
free_handle_memory(struct handle_list **base_ptr) {
	if (base_ptr != NULL) {
		struct handle_list *base = *base_ptr;
		struct handle_list *ptr = NULL, *ptr2 = NULL;

		if (base != NULL) {
			/* Free handle record memory */
			for (ptr = base; ptr != NULL; ) {
				ptr2 = ptr;
				ptr = ptr->next;
				if (ptr != NULL) {
					ptr->next = NULL;
				}
				free(ptr2);
			}
		}
		*base_ptr = NULL;
	}
}


/*
 * Add a handle to a linked list.
 * This is used to keep track of successful matches with various records.
 * Convert the handle to lower case.
 * Return 0 on success, non-zero on error.
 */
static int
add_handle(char *s, struct handle_list **base_ptr) {
	struct handle_list *newrec = NULL;
	struct handle_list *p = NULL;

	/* Sanity checking: */
	if ((s == NULL) || (base_ptr == NULL)) {
		print_syslog(LOG_ERR, "add_handle() passed NULL\n");
		return (-1);
	}

	/* Allocate and initialize memory: */
	newrec = (struct handle_list *) malloc(sizeof (struct handle_list));
	if (newrec == NULL) {
		print_syslog(LOG_ERR,
			"Cannot allocate memory for new handle; errno %d %s",
				errno, strerror(errno));
		return (-1);
	}
	memset(newrec, 0, sizeof (struct handle_list));

	/* Find end of list: */
	p = *base_ptr;
	if (p == NULL) { /* First item on the list */
		*base_ptr = newrec;

	} else { /* Add to end of list */
		/* Find list end: */
		while (p->next != NULL) {
			p = p->next;
		}

		/* Update previous link: */
		p->next = newrec;
	}

	/* Initialize new record: */
	lowercase_string(s);
	(void) strlcpy(newrec->handle, s, sizeof (newrec->handle));
	newrec->next = NULL;

	return (0);
}	/* add_handle() */


/*
 * Terminate a line with a NUL character.
 * (usually by overwriting a return or newline character).
 */
static void
strip_newline(char *s) {
	if (s == NULL) {
		print_syslog(LOG_ERR, "strip_newline() passed NULL\n");
		return;
	}

	while (!isendofline(*s)) {
		++s;
	}

	*s = '\0';
}


/*
 * Return a pointer to the idx-th token on a line,
 * where each token is separated by whitespace.
 * If none, return a zero-length string.
 */
static char *
line_ptr(int idx, char *buff) {
	int		i;
	static char	m[1] = "\0";

	if (buff == NULL) {
		print_syslog(LOG_ERR, "line_ptr() passed NULL\n");
		return (NULL);
	}

	--idx;

	while (isblank(*buff) && !isendofline(*buff)) {
		++buff;
	}

	if (isendofline(*buff)) {
		return (m);
	}

	for (i = 0; i < idx; ++i) {
		while (!isblank(*buff) && !isendofline(*buff)) {
			++buff;
		}
		if (isendofline(*buff)) {
			return (m);
		}
		while (isblank(*buff) && !isendofline(*buff)) {
			++buff;
		}
		if (isendofline(*buff)) {
			return (m);
		}
	}
	return (buff);
}


/*
 * Return a NUL-terminated substring for the idx-th token on a line,
 * where each token is separated by whitespace.
 */
static char *
line_entry(int idx, char *buff) {
	static char	b2[MAXLINE], *pb2 = NULL;
	int		i;

	if (buff == NULL) {
		print_syslog(LOG_ERR, "line_entry() passed NULL\n");
		return (NULL);
	}

	--idx;
	b2[0] = '\0';
	pb2 = b2;

	while (isblank(*buff) && !isendofline(*buff)) {
		++buff;
	}
	if (isendofline(*buff)) {
		return (b2);
	}

	for (i = 0; i < idx; ++i) {
		while (!isblank(*buff) && !isendofline(*buff)) {
			++buff;
		}
		if (isendofline(*buff)) {
			return (b2);
		}
		while (isblank(*buff) && !isendofline(*buff)) {
			++buff;
		}
		if (isendofline(*buff)) {
			return (b2);
		}
	}

	while (!isblank(*buff) && !isendofline(*buff)) {
		*pb2 = *buff;
		++buff;
		++pb2;
	}
	*pb2 = '\0';
	return (b2);
}


/*
 * Search for the current field type
 * and print it out.
 */
static void
print_field(char *cp, char *value, FILE *out_fd) {
	char	*title = NULL;

	if ((cp == NULL) || (value == NULL) || (out_fd == NULL)) {
		print_syslog(LOG_ERR, "print_field() passed NULL\n");
		return; /* error */
	}

	/* Domain Records */
	if (strcmp(cp, "name") == 0) {
		title = _("Name:");
	} else if (strcmp(cp, "org") == 0) {
		title = _("Organization:");
	} else if (strcmp(cp, "fqdn") == 0) {
		title = _("FQDN:");
	} else if (strcmp(cp, "arch") == 0) {
		title = _("Architecture:");
	} else if (strcmp(cp, "os") == 0) {
		title = _("Operating System:");
	} else if (strcmp(cp, "ip") == 0) {
		title = _("IP Address:");
	} else if (strcmp(cp, "contact") == 0) {
		title = _("Contact Handle:");
	} else if (strcmp(cp, "admin") == 0) {
		title = _("Administrative Contact:");
	} else if (strcmp(cp, "tech") == 0) {
		title = _("Technical Contact:");
	} else if (strcmp(cp, "country") == 0) {
		title = _("Country:");

	/* Host Records */
	} else if (strcmp(cp, "cname") == 0) {
		title = _("CNAME Record:");
	} else if (strcmp(cp, "mx") == 0) {
		title = _("MX Record:");
	} else if (strcmp(cp, "domain") == 0) {
		title = _("Domain Name:");
	} else if (strcmp(cp, "ns1") == 0) {
		title = _("Name Server 1:");
	} else if (strcmp(cp, "ns1ip") == 0) {
		title = _("NS1 IP address:");
	} else if (strcmp(cp, "ns2") == 0) {
		title = _("Name Server 2:");
	} else if (strcmp(cp, "ns2ip") == 0) {
		title = _("NS2 IP address:");
	} else if (strcmp(cp, "ns3") == 0) {
		title = _("Name Server 3:");
	} else if (strcmp(cp, "ns3ip") == 0) {
		title = _("NS3 IP address:");
	} else if (strcmp(cp, "ns4") == 0) {
		title = _("Name Server 4:");
	} else if (strcmp(cp, "ns4ip") == 0) {
		title = _("NS4 IP address:");
	} else if (strcmp(cp, "ns5") == 0) {
		title = _("Name Server 5:");
	} else if (strcmp(cp, "ns5ip") == 0) {
		title = _("NS5 IP address:");
	} else if (strcmp(cp, "ns6") == 0) {
		title = _("Name Server 6:");
	} else if (strcmp(cp, "ns6ip") == 0) {
		title = _("NS6 IP address:");
	} else if (strcmp(cp, "ns7") == 0) {
		title = _("Name Server 7:");
	} else if (strcmp(cp, "ns7ip") == 0) {
		title = _("NS7 IP address:");
	} else if (strcmp(cp, "ns8") == 0) {
		title = _("Name Server 8:");
	} else if (strcmp(cp, "ns8ip") == 0) {
		title = _("NS8 IP address:");
	} else if (strcmp(cp, "ns9") == 0) {
		title = _("Name Server 9:");
	} else if (strcmp(cp, "ns9ip") == 0) {
		title = _("NS9 IP address:");

	/* Person Records */
	} else if (strcmp(cp, "email") == 0) {
		title = _("Email address:");
	} else if (strcmp(cp, "contact") == 0) {
		title = _("Contact:");
	} else if (strcmp(cp, "address") == 0) {
		title = _("Address:");
	} else if (strcmp(cp, "phone") == 0) {
		title = _("Telephone:");
	} else if (strcmp(cp, "fax") == 0) {
		title = _("FAX No:");
	} else if (strcmp(cp, "created") == 0) {
		title = _("Created:");
	} else if (strcmp(cp, "updated") == 0) {
		title = _("Last Updated:");

	/* Network Records */
	} else if (strcmp(cp, "net") == 0) {
		title = _("Network:");
	} else if (strcmp(cp, "as") == 0) {
		title = _("AS:"); /* Autonomous System (in BGP protocol) */
	}

	if (title != NULL) { /* found--print it */
		(void) fprintf(out_fd, "%-25s %s", title, value);
	}
}




/*
 * Perform the requested whois operation.
 * Read input from stdin and write to out_fd.
 * Return 0 on success, non-zero on error.
 */
int
do_whois(FILE *in_fd, FILE *out_fd, char *ident) {
	char			client_input[MAXLINE] = "";
	char			raw_hdr[MAXLINE];
	char			cooked_hdr[MAXLINE];
	char			raw_line[MAXLINE];
	char			value[MAXLINE];
	char			search_hdr[MAXLINE];
	char			read_line[MAXLINE];
	FILE			*db_fp = NULL;
	int			record_count = 0;
	int			rlen = 0; /* response len (#requests output) */
	struct handle_list	*base = NULL, *ptr = NULL;


	/* Sanity check */
	if ((in_fd == NULL) || (out_fd == NULL)) {
		log_to_daemonlog(fileno(in_fd), client_input, ident,
			HTTP_INTERNAL_SERVER_ERROR, rlen);
		print_syslog(LOG_ERR, "do_whois() passed NULL\n");
		return (1); /* error */
	}

	/* Remove previous results, if any: */
	base = NULL;

	/* Print header */
	(void) fputs(SWHOISD_BANNER "\r\n", out_fd);
	(void) print_file_if_present(SWHOISD_hdr, out_fd);

	/* Read input from whois client */
	if (fgets(client_input, sizeof (client_input), in_fd) == NULL) {
#if ! defined WILDCARD_SEARCH_ALLOWED
		log_to_daemonlog(fileno(in_fd), client_input, ident,
				HTTP_BAD_REQUEST, rlen);
		(void) fprintf(out_fd, _("Empty request"), SWHOISD_CONF);
		(void) fputs("\r\n\r\n", out_fd); /* '\r' can't be in gettext */
		return (-1);
#endif
	}


	strip_newline(client_input);
	lowercase_string(client_input);

	/* Open database file for searching */
	db_fp = fopen(SWHOISD_CONF, "r");
	if (db_fp == NULL) {
		print_syslog(LOG_ERR, "Failed to open database file %s.",
			SWHOISD_CONF);
		(void) fprintf(out_fd, _("Failed to open database file %s."),
			SWHOISD_CONF);
		(void) fputs("\r\n\r\n", out_fd); /* '\r' can't be in gettext */
		log_to_daemonlog(fileno(in_fd), client_input, ident,
				HTTP_NO_CONTENT, rlen);
		return (-1);
	}

	(void) fprintf(out_fd, _("Searching for %s."), client_input);
	(void) fputs("  ", out_fd);
	(void) fflush(out_fd);

#ifndef ALLOW_DENIAL_OF_SERVICE
	/* Help prevent DOS by delaying response for a short time */
	(void) sleep(delay_time);
#endif

	/*
	 * Search database file
	 */
	while (fgets(raw_hdr, MAXLINE, db_fp) !=  NULL) {
		/* skip comment line, if any: */
		int i = 0;

		while (isblank(raw_hdr[i])) {
			++i;
		}
		if (raw_hdr[i] == '#') {
			continue;
		}

		/* Copy line and convert to lower case for comparison */
		strip_newline(raw_hdr);
		(void) strlcpy(cooked_hdr, line_entry(1, raw_hdr),
		sizeof (cooked_hdr));
		lowercase_string(cooked_hdr);

		if (strcmp(line_entry(3, raw_hdr), "{") == 0) {
			/* Dump ALL records in the database: */
			if (strcmp(client_input, "*") == 0) {
#ifdef WILDCARD_SEARCH_ALLOWED
				if (add_handle(cooked_hdr, &base) != 0) {
					print_syslog(LOG_ERR,
						"add_handle failed");
					log_to_daemonlog(fileno(in_fd),
						client_input, ident,
						HTTP_INTERNAL_SERVER_ERROR,
						rlen);
					free_handle_memory(&base);
					fclose(db_fp);
					return (-1);
				}
				++record_count;
				continue;
#else
				(void) fprintf(out_fd,
					_("Wildcard (*) requests not allowed "
					"on this server."),
					client_input);
				log_to_daemonlog(fileno(in_fd), client_input,
					ident, HTTP_FORBIDDEN, rlen);
				free_handle_memory(&base);
				fclose(db_fp);
				return (-1);
#endif
			}

			if (strcmp(client_input, cooked_hdr) == 0) {
				if (add_handle(cooked_hdr, &base) != 0) {
					print_syslog(LOG_ERR,
						"add_handle failed");
					log_to_daemonlog(fileno(in_fd),
						client_input, ident,
						HTTP_INTERNAL_SERVER_ERROR,
						rlen);
					free_handle_memory(&base);
					fclose(db_fp);
					return (-1);
				}
				++record_count;
			} else if ((client_input[0] == '!') &&
					(strcmp(&client_input[1],
							cooked_hdr) == 0)) {
				/* Lines starting with ! are handle requests */
				if (add_handle(cooked_hdr, &base) != 0) {
					print_syslog(LOG_ERR,
						"add_handle failed");
					log_to_daemonlog(fileno(in_fd),
						client_input, ident,
						HTTP_INTERNAL_SERVER_ERROR,
						rlen);
					free_handle_memory(&base);
					fclose(db_fp);
					return (-1);
				}
				++record_count;
			} else {
				do {
					if (fgets(raw_line, MAXLINE,
							db_fp) == NULL) {
						break;
					}

					(void) strlcpy(value,
						line_entry(1, raw_line),
						sizeof (value));

					/*
					 * If a field starts with '!',
					 * it is searchable--so search it:
					 */
					if (value[0] == '!') {
						(void) strlcpy(search_hdr,
							line_ptr(2, raw_line),
						sizeof (search_hdr));
						strip_newline(search_hdr);
						lowercase_string(search_hdr);

						if (strstr(search_hdr,
								client_input)
								!= NULL) {
						 if (add_handle(
								cooked_hdr,
								&base) != 0) {
						  print_syslog(LOG_ERR,
							"add_handle failed");
						  log_to_daemonlog(
						    fileno(in_fd),
						    client_input, ident,
						    HTTP_INTERNAL_SERVER_ERROR,
						    rlen);
						  free_handle_memory(
							&base);
						  fclose(db_fp);
						  return (-1);
						 }
						 ++record_count;
						 break;
						}
					}
				} while (strcmp(line_entry(1, raw_line), "}"));
			}
		}	/* if */
	}	/* while */

	if (record_count == 0) {
		(void) fputs("\r\n", out_fd); /* '\r' can't be in gettext */
		(void) fprintf(out_fd, _("No records matching %s found."),
			client_input);
		(void) fputs("\r\n\r\n", out_fd); /* '\r' can't be in gettext */
		log_to_daemonlog(fileno(in_fd), client_input,
			ident, HTTP_NOT_FOUND, rlen);
		free_handle_memory(&base);
		fclose(db_fp);
		return (-1);
	}

	(void) fprintf(out_fd, _("Found %1$d record(s) matching %2$s."),
			record_count, client_input);
	(void) fputs("\r\n\r\n", out_fd); /* '\r' can't be in I18N gettext */

	/* Reread database file from beginning */
	(void) fseek(db_fp, 0L, SEEK_SET);
	if (db_fp == NULL) {
		(void) fputs("\r\n", out_fd); /* '\r' can't be in gettext */
		(void) fprintf(out_fd,
			_("Failed to fseek(0) database file %s."),
			SWHOISD_CONF);
		(void) fputs("\r\n", out_fd); /* '\r' can't be in gettext */
		log_to_daemonlog(fileno(in_fd), client_input,
			ident, HTTP_INTERNAL_SERVER_ERROR, rlen);
		free_handle_memory(&base);
		fclose(db_fp);
		return (-1);
	}

	/*
	 * Print out the records we found.
	 */
	for (ptr = base; ptr != NULL; ) {
		if (fgets(read_line, MAXLINE, db_fp) == NULL) {
			break;
		}

		strip_newline(read_line);

		if (strcmp(line_entry(3, read_line), "{") == 0) {
			(void) strlcpy(search_hdr, line_entry(1, read_line),
				sizeof (search_hdr));
			lowercase_string(search_hdr);

			if (strcmp(search_hdr, ptr->handle) == 0) {
				char *record_title = NULL;

				(void) strlcpy(raw_hdr,
					line_entry(2, read_line),
					sizeof (raw_hdr));
				lowercase_string(raw_hdr);

				/* Print out the header for a record */
				if (strcmp(raw_hdr, "hr") == 0) {
					record_title = _("Host:");
				} else if (strcmp(raw_hdr, "dr") == 0) {
					record_title = _("Domain:");
				} else if (strcmp(raw_hdr, "pr") == 0) {
					record_title = _("Person:");
				} else if (strcmp(raw_hdr, "nr") == 0) {
					record_title = _("Network:");
				}
				if (record_title != NULL) {
					++rlen;
					(void) fprintf(out_fd, "%-25s ",
						record_title);
					(void) fprintf(out_fd, _("(Handle %s)"),
						line_entry(1, read_line));
					(void) fputs("\r\n", out_fd);
				}

				/* Read fields in a record and print it out */
				do {
					char *cp = NULL;
					if (fgets(cooked_hdr, MAXLINE, db_fp)
							== NULL)
						break;
					(void) strlcpy(raw_line,
						line_entry(1, cooked_hdr),
						sizeof (raw_line));
					(void) strlcpy(value,
						line_ptr(2, cooked_hdr),
						sizeof (value));


					/*
					* Flush '!' (from file) and
					* convert to lower case for comparison.
					*/
					if (raw_line[0] == '!') {
						cp = &raw_line[1];
					} else {
						cp = raw_line;
					}
					lowercase_string(cp);

					/* Print the current field type */
					print_field(cp, value, out_fd);
				} while (strcmp(line_entry(1, cooked_hdr),
						"}"));

				(void) fputs("\r\n", out_fd);
				ptr = ptr->next;
			}
		}	/* if */
	}	/* for */

	/* Print footer */
	if ((print_file_if_present(SWHOISD_FOOTER, out_fd)) != 0) {
		/* print default footer if footer file isn't present */
		(void) fprintf(out_fd, _(
			"To single out one record, look it up with \"!XXX\","
			" where XXX is the handle"));
		(void) fputs("\r\n", out_fd); /* '\r' can't be in gettext */
	}
	(void) fputs("\r\n", out_fd);
	(void) fflush(out_fd);

	/* Cleanup search: */
	free_handle_memory(&base);
	(void) fclose(db_fp);

	/* Log successful request: */
	log_to_daemonlog(fileno(in_fd), client_input, ident, HTTP_OK, rlen);

	return (0);
}	/* do_whois */
