/*
 * rfc1413.c
 *
 * Copyright  2001 Dan Anderson <dan@drydog.com> All rights reserved.
 * See file COPYING for licensing revisions.
 *
 * rfc1413() speaks a common subset of the RFC 1413, AUTH, TAP, IDENT and RFC
 * 931, 1413 protocols.
 * It queries an RFC 1413 etc.-compatible daemon on a remote
 * host to look up the owner of a connection. The information should not be
 * used for authentication purposes. This routine intercepts alarm signals.
 *
 * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
 *
 * Modified by Dan Anderson, San Diego, California, July 2001:
 * Added more error messages.
 * Fixed bug in checking fgets() return value (hint: it's not a integer)
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char rcsid[] = "@(#) $Id: rfc1431.c,v 1.3 2001/07/04 06:14:22 dan Exp $";
/* static char sccsid[] = "@(#) rfc931.c 1.10 95/01/02 16:11:34"; */
#endif

#include "config.h"
#include "in.swhoisd.h"

/* System libraries. */

#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>

/* Local stuff. */

#define	RFC1413_PORT		113	/* Well-known port */
#define	ANY_PORT		0	/* Any old port will do */
#define	RFC1413_TIMEOUT		30	/* seconds */
#define	UNKNOWN_RESPONSE	"Unknown"
#define	RFC1413_USERLEN		512
#define	RFC1413_MAXDATA		1024	/* RFC 1413 recommends 1000 */
#define	MAX_USERID		256	/* Arbitrary, < RFC_USERLEN */
#define	INVALID_PORT		65535

int	rfc1413_timeout = RFC1413_TIMEOUT; /* Global so it can be changed */

static jmp_buf timebuf;

/* timeout - handle SIGALARM timeouts */
static void
timeout(int sig) {
	longjmp(timebuf, sig);
}


/*
 * rfc1413 - return remote user name, given socket structures
 * This is the entry point function.
 * Return 0 if OK, non-zero on error.
 */
int
rfc1413(struct sockaddr_in *rmt_sin, struct sockaddr_in *our_sin,
		char *dest, unsigned dest_len) {
	in_port_t		rmt_port, our_port;
	/* These can't be in_port_t (unsigned short) due to sscanf(): */
	unsigned int		rmt_port_reply, our_port_reply;
	struct sockaddr_in	rmt_query_sin, our_query_sin;
	char			user[MAX_USERID];
	char			buffer[RFC1413_USERLEN];
	char			*cp = NULL;
	char			*result = UNKNOWN_RESPONSE;
	FILE			*fp_in = NULL;
	FILE			*fp_out = NULL;
	struct sigaction	sa;
	int			len_read, r;
	int			sock;

	/* print_syslog(LOG_ERR, "DEBUG: rfc1431 called"); */

	/* Sanity checking: */
	if (rmt_sin == NULL) {
		print_syslog(LOG_ERR, "NULL rmt_sin passed to rfc1413()\n");
		return (-1);
	} else if (our_sin == NULL) {
		print_syslog(LOG_ERR, "NULL our_sin passed to rfc1413()\n");
		return (-1);
	} else if (dest == NULL) {
		print_syslog(LOG_ERR, "NULL dest passed to rfc1413()\n");
		return (-1);
	} else if (dest_len == 0) {
		print_syslog(LOG_ERR, "0 dest_len passed to rfc1413()\n");
		return (-1);
	}

	/*
	 * Use one unbuffered stdio stream for writing to and for reading from
	 * the RFC1413 etc. server. This is unbuffered because of a bug in the
	 * SunOS 4.1.x stdio library. The bug may live in other stdio
	 * implementations, too. When we use a single, buffered, bidirectional
	 * stdio stream ("r+" or "w+" mode) we read our own output. Such
	 * behavior would make sense with resources that support random-access
	 * operations, but not with sockets.
	 */
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		print_syslog(LOG_ERR, "socket() failed; errno %d %s\n",
				errno, strerror(errno));
		return (-1);

	}

	if ((fp_in = fdopen(sock, "r")) == 0) {
		print_syslog(LOG_ERR,
			"fdopen(%d, \"r\") failed; errno %d %s\n",
			sock, errno, strerror(errno));
		(void) close(sock);
	}

	if ((fp_out = fdopen(sock, "w")) == 0) {
		print_syslog(LOG_ERR,
			"fdopen(%d, \"w\") failed; errno %d %s\n",
			sock, errno, strerror(errno));
		(void) close(sock);
	}

	/* Disable buffering for the output socket: */
	(void) setvbuf(fp_out, NULL, _IONBF, 0);

	/*
	 * Set up a timer so we won't get stuck while waiting for the
	 * server.
	 */
	if (setjmp(timebuf) != 0) {
		print_syslog(LOG_ERR, "setjmp() failed; errno %d %s\n",
				errno, strerror(errno));
		(void) fclose(fp_in);
		(void) fclose(fp_out);
		return (-1);
	}

	/* Setup alarm signal handler: */
	(void) memset(&sa, 0, sizeof (sa));
	sa.sa_handler = timeout;
	(void) sigaction(SIGALRM, &sa, NULL);
	/* don't use signal()--it's not as "reliable" */

	(void) alarm(rfc1413_timeout);

	/*
	 * Bind the local and remote ends of the query socket
	 * to the same IP addresses as the connection under
	 * investigation. We go through all this trouble
	 * because the local or remote system might have more
	 * than one network address. The RFC1413 etc. client
	 * sends only port numbers; the server takes the IP
	 * addresses from the query socket.
	 */
	our_query_sin = *our_sin;
	our_query_sin.sin_port = htons(ANY_PORT);
	rmt_query_sin = *rmt_sin;
	rmt_query_sin.sin_port = htons(RFC1413_PORT);

	if (bind(sock, (struct sockaddr *)&our_query_sin,
			sizeof (our_query_sin)) == -1) {
		print_syslog(LOG_ERR, "bind(%d) failed; errno %d %s\n",
			sock, errno, strerror(errno));
		(void) alarm(0); /* disarm */
		(void) fclose(fp_in);
		(void) fclose(fp_out);
		return (-1);
	}

	if (connect(sock, (struct sockaddr *)&rmt_query_sin,
			sizeof (rmt_query_sin)) == -1) {
		/*
		 * ECONNREFUSED errors from connect() usually imply the
		 * machine doesn't support the service. This is common.
		 */
		if (errno != ECONNREFUSED) {
			print_syslog(LOG_ERR,
				"connect(%d) failed; errno %d %s\n",
				sock, errno, strerror(errno));
		/* }else{print_syslog(LOG_ERR, "DEBUG: connection refused");*/
		}
		(void) alarm(0); /* disarm */
		(void) fclose(fp_in);
		(void) fclose(fp_out);
		return (0); /* OK */
	}

	/*
	 * Send query to server. Neglect the risk that a
	 * 13-byte write would have to be fragmented by the
	 * local system and cause trouble with buggy System V
	 * stdio libraries.
	 */
	rmt_port = ntohs(rmt_sin->sin_port);
	our_port = ntohs(our_sin->sin_port);
	if ((rmt_port == INVALID_PORT) || (rmt_port == INVALID_PORT)) {
		print_syslog(LOG_ERR, "Invalid port: remote %u, local %d\n",
				rmt_port, our_port);
		(void) alarm(0); /* disarm */
		(void) fclose(fp_in);
		(void) fclose(fp_out);
	}
	(void) fprintf(fp_out, "%u,%u\r\n", rmt_port, our_port);
	(void) fflush(fp_out);
	/* print_syslog(LOG_ERR, "DEBUG: query %u,%u", rmt_port, our_port); */

	/*
	 * Read response from server. Use fgets()/sscanf() so
	 * we can work around System V stdio libraries that
	 * incorrectly assume EOF when a read from a socket
	 * returns less than requested.
	 */
	(void) fgets(buffer, sizeof (buffer), fp_in);
	if (ferror(fp_in) || feof(fp_in)) {
		if (feof(fp_in)) {
			print_syslog(LOG_ERR, "fgets() failed; returned EOF\n");
		} else {
			print_syslog(LOG_ERR, "fgets() failed; errno %d %s \n",
				errno, strerror(errno));
		}
		(void) alarm(0); /* disarm */
		(void) fclose(fp_in);
		(void) fclose(fp_out);
		return (-1);
	}

	(void) alarm(0); /* disarm */
	/* print_syslog(LOG_ERR, "DEBUG: response: %s", buffer); */

	/* Parse reply: */
	if ((r = sscanf(buffer, "%u , %u : USERID :%*[^:]:%255s",
			&rmt_port_reply, &our_port_reply, user)) != 3) {
		print_syslog(LOG_ERR, "scanf() failed; returned %d;"
			" buffer: %s\n", r, buffer);
		(void) fclose(fp_in);
		return (-1);
	}

	/* Verify ports match our request: */
	if ((rmt_port_reply == rmt_port) && (our_port_reply == our_port)) {
		/*
		 * Success!
		 * Strip trailing carriage return.  It is part
		 * of the protocol, not part of the data.
		 */
		if (cp = strchr(user, '\r')) {
			*cp = 0;
		}
		result = user;
	}

	(void) fclose(fp_in);
	(void) fclose(fp_out);

	/* Save answer: */
	(void) strlcpy(dest, result, dest_len);
	return (0); /* OK */
}	/* rfc1413() */
