/*
 * sysutils.c
 *
 * Copyright  2001 Dan Anderson <dan@drydog.com> All rights reserved.
 * See file COPYING for licensing terms.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char rcsid[] = "@(#) $Id: sysutils.c,v 1.4 2004/01/04 05:34:21 dan Exp $";
#endif

#include "config.h"
#include "in.swhoisd.h"

#include <sys/types.h>
#include <stdio.h>
#include <ctype.h>		/* is*() */
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>		/* sigaction() and friends */
#include <unistd.h>		/* fork(), setsid() */
#include <fcntl.h>		/* open() */
#include <syslog.h>		/* syslog(), LOG_ERR */
#include <pthread.h>		/* pthread_*() */
#include <stdarg.h>		/* ANSI C va_*() */
#include <grp.h>		/* getgrnam_r() */
#include <pwd.h>		/* getpwnam_r() */

/*
 * Socket stuff
 */
#include <sys/stat.h>		/* S_IFSOCK */
#include <sys/socket.h>		/* getpeername(), isfdtype() */
#include <netdb.h>		/* gethostbyaddr() */



/*
 * Constants
 */
#define	MAXFD		255
#define	LISTENQ		1024
#define	PWD_BUFFER_LEN	1024
#define	GRP_BUFFER_LEN	1024
#define	WHOIS_SERVICE	"whois"		/* 43/tcp */
#define	WHOIS_PORT	"43"		/* whois 43/tcp */

static FILE *swhois_log = NULL;		/* Daemon log */

static char *dash_string = "-";

/*
 * Parameters to pass to a new thread.
 * This must be allocated in advance (NOT placed on the stack) by the
 * parent thread and freed after copying by the child thread.
 */
struct whois_thread_parameters {
	int			connfd;
	struct sockaddr_in	*client_addr;
	};

/*
 * Close all files >= start_fd.
 */
static void
closeall(int start_fd) {
	int i;
	for (i = start_fd; i < MAXFD; ++i) {
		(void) close(i);
	}
	errno = 0;	/* clear probable errno from above */
}


#if __sun
/*
 * Return 1 if the fd is of the specified type, 0 if not, -1 on error.
 * Present in Linux (sys/socket.h), but not Solaris.
 * See W. Richard Stevens _UNIX Network Programming_, 2d ed. v.1, p. 82.
 */
int
isfdtype(int fd, int fdtype) {
	struct stat buf;

	if (fstat(fd, &buf) == -1) {
		print_syslog(LOG_ERR, "fstat(%d) failed; errno %d %s\n",
			fd, errno, strerror(errno));
		return (-1);
	}

	/* See if bit in there after masking and return 0 or 1. */
	if ((buf.st_mode & S_IFMT) == fdtype) {
		return (1);
	} else {
		return (0);
	}
}
#endif


/*
 * Daemonize ourself by detaching from the controlling
 * terminal and process and go into the background.
 * Open syslog.
 *
 * Return zero on success, non-zero on failure.
 * Believed to work on all Posix systems.
 *
 * See W. Richard Stevens _UNIX Network Programming_, 2d ed. v.1, p. 336.
 */
int
daemonize(void) {
	/* Signals to ignore: */
	static	signals_to_ignore[] = {SIGHUP, SIGTTOU, SIGTTIN, SIGTSTP, 0};
	int	sig;


	/* Fork once to the parent thinks we exited . . . */
	switch (fork()) {
	case 0: /* child */
		break;
	case -1: /* error */
		print_syslog(LOG_ERR|LOG_DAEMON,
			"fork(1) failed; errno %d %s\n",
			errno, strerror(errno));
		return (-1);
	default: /* parent */
		exit(0);
	}

	/* . . . become session leader . . . */
	(void) setsid();

	/*
	 * . . . ignore signals . . .
	 * Ignore the SIGT* terminal stop signals (BSD).
	 * System V: SIGHUP immune from pgrp leader death.
	 */
	for (sig = 0; (signals_to_ignore[sig] != 0); ++sig) {
		struct sigaction sa;
		(void) memset(&sa, 0, sizeof (sa));
		sa.sa_handler = SIG_IGN;
		(void) sigaction(signals_to_ignore[sig], &sa, NULL);
		/* Don't use signal()--it's not as "reliable" */
	}

	/*
	 * . . . and fork again.
	 * This disallows automatically acquiring a controlling tty.
	 */
	switch (fork()) {
	case 0: /* child */
		break;
	case -1: /* error */
		print_syslog(LOG_ERR|LOG_DAEMON,
			"fork(2) failed; errno %d %s\n",
			errno, strerror(errno));
		return (-1);
	default: /* parent */
		exit(0);
	}

	/* Close all open files and open dummy STDIN, STDOUT, STDERR; */
	closeall(0); /* not POSIX */
	(void) open("/dev/null", O_RDWR);
	(void) dup(0);
	(void) dup(0);

	return (0); /* OK */
}


/*
 * Get the current time/date and format and return the result in a string.
 * Use the HTTP's Common Logfile Format (CLF) standard date stamp:
 *	"%d/%b/%Y:%H:%M:%S"  For example: "26/Jun/2001:19:15:44 -0700"
 * Offset is hhmm +/- of UTC (e.g., -0700).  CLF has no DST adjustment.
 * If time is not known, return an error string (not NULL).
 */
char *
get_timestamp(char *buffer, int buffer_size) {
	time_t		curtime = 0L;
	struct tm	*loctime = NULL;
	struct tm	tm_buf;
	extern time_t	timezone; /* not set until AFTER localtime_r called */
	time_t		timezone_minutes;

	if ((buffer == NULL) || (buffer_size <= 0)) {
		return (NULL);
	}

	/* Get the current time. */
	(void) time(&curtime);

	/* Convert it to local time representation. */
	(void) memset(&tm_buf, 0, sizeof (tm_buf));
	loctime = localtime_r(&curtime, &tm_buf);

	/* Immediately save global variable and adjust seconds to minutes: */
	timezone_minutes	= timezone / 60;

	/* Return the date and time in the standard format. */
	if (strftime(buffer, buffer_size, TIMESTAMP_FORMAT,
			loctime) == 0) {
		return ("Time unknown");
	} else {

		/* Append timezone offset */
		int	len_used = strlen(buffer);
		if (timezone_minutes < 0) { /* positive offset from UTM */
			(void) snprintf(&buffer[len_used],
				sizeof (buffer) - len_used,
				" +%02ld%02ld", timezone_minutes / 60,
					timezone_minutes % 60);
		} else { /* negative offset from UTM */
			timezone_minutes = abs(timezone_minutes);
			(void) snprintf(&buffer[len_used],
				sizeof (buffer) - len_used,
				" -%02ld%02ld", timezone_minutes / 60,
					timezone_minutes % 60);
		}
		return (buffer);
	}
}


/*
 * Get the "nobody" (or if none, SCO's "nouser") UID value for this system.
 * Use a default if not found.
 *
 * For Sun, _POSIX_PTHREAD_SEMANTICS must be defined
 * so Sun's nonstandard getpwnam_r() prototype won't be used.
 */
uid_t
get_nobody_uid(void) {
	char		pwd_buff[PWD_BUFFER_LEN];
	int		e;
	uid_t		uid;
	struct passwd	pwd, *pwd_ptr = NULL;

	pwd.pw_uid = MYUID;
	if ((e = getpwnam_r("nobody", &pwd, pwd_buff,
			sizeof (pwd_buff), &pwd_ptr)) == 0) {
		uid = pwd.pw_uid;

	} else if ((e = getpwnam_r("nouser", &pwd, pwd_buff,
				sizeof (pwd_buff),
				&pwd_ptr)) == 0) {
		/* error--use SCO ODT 3's "nouser" */
		uid = pwd.pw_uid;

	} else { /* error--use default for this system */
		print_syslog(LOG_ERR, "getgrname_r() failed;"
			" errno %d %s\n", e, strerror(e));
		uid = MYUID;
	}

	return (uid);
}	/* get_nobody_uid() */


/*
 * Get the "nogroup" (or if none, "nobody") GID value for this system.
 * Use a default if not found.
 *
 * For Sun, _POSIX_PTHREAD_SEMANTICS must be defined
 * so Sun's nonstandard getgrnam_r() prototype won't be used.
 */
gid_t
get_nogroup_gid(void) {
	char		grp_buff[GRP_BUFFER_LEN];
	int		e;
	gid_t		gid;
	struct group	grp, *grp_ptr = NULL;

	grp.gr_gid = MYGID;
	if ((e = getgrnam_r("nogroup", &grp, grp_buff,
			sizeof (grp_buff), &grp_ptr)) == 0) {
		gid = grp.gr_gid;

	} else if ((e = getgrnam_r("nobody", &grp, grp_buff,
				sizeof (grp_buff),
				&grp_ptr)) == 0) {
		/* error--try "nobody" */
		gid = grp.gr_gid;

	} else { /* error--use UID nobody's GID */
		char		pwd_buff[PWD_BUFFER_LEN];
		struct passwd	pwd, *pwd_ptr = NULL;

		pwd.pw_uid = MYUID;
		if ((e = getpwnam_r("nobody", &pwd, pwd_buff,
				sizeof (pwd_buff), &pwd_ptr)) == 0) {
			gid = pwd.pw_gid;

		} else { /* error--use default for this system */
			print_syslog(LOG_ERR, "getgrname_r() failed;"
				" errno %d %s\n", e, strerror(e));
			gid = MYGID;
		}
	}

	return (gid);
}	/* get_nogroup_gid() */


/* sighup_signal_handler
 *
 * Handle the SIGHUP signal.
 * Reopen log files (and any other files, including configuration files,
 * if necessary).
 */
static void sighup_signal_handler(int signo)
{
	if (enable_logging) {
		if (swhois_log != NULL) {
			fclose(swhois_log);
			swhois_log = NULL;
		}

		/* Open Daemon log file--must have correct ownership */
		swhois_log = fopen(DAEMON_LOG_FILENAME, "a");
		if (swhois_log == NULL) {
			print_syslog(LOG_ERR,
				"SIGHUP: fopen(swhois_log) failed;"
				" errno %d %s\n", errno, strerror(errno));
		}
	}
}


/*
 * Initialization performed after daemonizing.
 * Also called for inetd mode (no daemonizing).
 * Open syslog.
 * Set umask to 022 (0 is a security risk).
 *
 * Return zero on success, non-zero on failure.
 *
 * See W. Richard Stevens _UNIX Network Programming_, 2d ed. v.1.
 */
int
initialize1(void) {
	int	r;

	/* cd to root to prevent problems unmounting a filesystem: */
	(void) chdir("/");

	/* Set our file mode creation mask to a safe value: */
	(void) umask(022);

	/* Open syslog (syslog.conf value "*.daemon") */
	openlog(MYNAME, LOG_PID|LOG_CONS, LOG_DAEMON);

}	/* initialize1() */


/*
 * Initialization performed after sockets are open.
 * If UID or GID are root, lower to non-root values.
 * Open the application log.
 *
 * Return zero on success, non-zero on failure.
 */
int
initialize2(void) {
	int	r;
	struct sigaction sa;

	/* For safety and security, don't run as root: */
	if (getgid() == 0) { /* root group (must be set before setuid()) */
		if (setgid(desired_gid) == -1) {
			print_syslog(LOG_ERR, "setgid() failed; errno %d %s\n",
				errno, strerror(errno));
		}
	}

	if (getuid() == 0) { /* root */
		if (setuid(desired_uid) == -1) {
			print_syslog(LOG_ERR, "setuid() failed; errno %d %s\n",
				errno, strerror(errno));
		}
	}

	/* Open Daemon log file--must have correct ownership */
	if (enable_logging) {
		swhois_log = fopen(DAEMON_LOG_FILENAME, "a");
		if (swhois_log == NULL) {
			print_syslog(LOG_ERR, "fopen(swhois_log) failed;"
				" errno %d %s\n", errno, strerror(errno));
			return (-1); /* error */
		}
	}

	/* Line-buffer the log file: */
	r = setvbuf(swhois_log, NULL, _IOLBF, 0);
	if (r == -1) {
		print_syslog(LOG_ERR,
			"setvbuf(swhois_log) failed; errno %d %s\n",
			errno, strerror(errno));
		return (-1); /* error */
	}

	/* Set up SIGHUP signal handler (which reopens the log file) */
	(void) memset(&sa, 0, sizeof (sa));
	sa.sa_handler = sighup_signal_handler;
	(void) sigaction(SIGHUP, &sa, NULL);
	/* Don't use signal()--it's not as "reliable" */

	return (0); /* OK */
}	/* initialize2() */


/*
 * Print fprint-like to the Daemon log file
 */
/*PRINTFLIKE1*/
void
print_daemonlog(char *fmt, ...) {
	va_list ap;
	va_start(ap, fmt);
	if (enable_logging && (swhois_log != NULL)) {
		(void) vfprintf(swhois_log, fmt, ap);
	}
	va_end(ap);
}	/* print_daemonlog() */


/*
 * Print fprint-like to syslog
 */
/*PRINTFLIKE2*/
void
print_syslog(int priority, char *fmt, ...) {
	char	buff[MAXLINE];
	va_list ap;
	va_start(ap, fmt);
	(void) vsnprintf(buff, sizeof (buff) - 1, fmt, ap);
	syslog(priority, buff);
	va_end(ap);
}	/* print_syslog() */


/*
 * Add an entry to the daemon log.
 * Input:
 *	request is the whois request.  NULL if none or unknown.
 *	status  is the HTTP status code (usually HTTP_OK)
 *	len	is the number of records retrieved (0 on error)
 *
 * Emulate the HTTP's Common Logfile Format (CLF).  That is:
 *	host ident - [date] "GET request" returncode size
 * Unknown fields are "-"
 * For example:
 *	129.107.83.23 fred - [26/Jun/2001:19:47:24 -0700] - - -
 * This should make it relatively easy for third-party web analyzing
 * tools, such as analog, to parse and analyze the swhoisd logfile.
 *
 * Note:
 * len should really be the response length in bytes,
 * but let's just "wing it" here and give the
 * number of records found--at least for now.
 */
void
log_to_daemonlog(int sock, char *request, char *ident, int status, int len) {
	char	timestring[MAXLINE];
	char	get_request[MAXLINE];

	if (enable_logging && (swhois_log != NULL)) {
		char *host = NULL;

		if (sock < 0) {
			host = dash_string; /* unknown */
		} else {
			host = get_client_hostname(sock);
			if ((host == NULL) || (host[0] == '\0')) {
				host = dash_string; /* unknown */
			}
		}

		(void) get_timestamp(timestring, sizeof (timestring));

		if ((ident == NULL) || (ident[0] == '\0')) {
			ident = dash_string; /* unknown */
		}

		if ((request == NULL) || (request[0] == '\0')) {
			(void) strlcpy(get_request, dash_string,
				sizeof (get_request)); /* unknown */
		} else {
			/* Fake a HTTP GET request: */
			(void) snprintf(get_request, sizeof (get_request),
				"GET %s", request);
		}

		print_daemonlog("%s %s - [%s] \"%s\" %d %d\n",
			host, ident, timestring, get_request, status, len);
	}
}


/*
 * Perform cleanup functions, such as closing logs, sockets, and files.
 * Done once per process (before exit()).
 */
void
cleanup(void) {
	if (enable_logging && (swhois_log != NULL)) {
		(void) fclose(swhois_log);
	}
	closelog();
}


#ifndef INET_ONLY
/*
 * Return 1 if a socket, 0 if not a socket, -1 on error.
 */
int
issocket(int sockfd) {
	return (isfdtype(sockfd, S_IFSOCK));
}
#endif


/*
 * Get the hostname (or IP address) of the connecting client.
 * When running in inetd mode, pass stdin (0).
 * Return the hostname.
 * Return NULL on error.
 */
#ifdef HAVE_GETPEERNAME
char *
get_client_hostname(int sockfd) {
	struct sockaddr_in sin;
	int len = sizeof (sin);
	struct hostent *host = NULL;

	if (sockfd >= 0) {
		if (getpeername(sockfd, (struct sockaddr *) &sin, &len)
				!= -1) {
			if ((host = gethostbyaddr((char *) &sin.sin_addr,
					sizeof (sin.sin_addr),
					AF_INET)) != NULL) {
				/* printf("remote host %s\n", host->h_name); */
				return (host->h_name);
			} else {
				print_syslog(LOG_ERR, "gethostbyaddr() failed;"
					" errno %d %s\n",
					errno, strerror(errno));
			}
		} else {
			print_syslog(LOG_ERR, "getpeername(%d) failed;"
				" errno %d %s\n",
				sockfd, errno, strerror(errno));
		}
	}
	return (NULL); /* invalid sockfd--hostname unknown */
}	/* get_client_hostname() */
#else
#error "get_client_hostname() not implemented yet without getpeername()"
#endif


#ifndef INET_ONLY
/*
 * This function does setup for clients.
 * It setups and makes a socket connection to a server. That is:
 * 1.	Create a socket,
 * 2.	connect to a server's host and well-known port.
 *
 * See W. Richard Stevens _UNIX Network Programming_, 2d ed. v.1, p. 285.
 */
#ifdef HAVE_GETADDRINFO
int
tcp_connect(const char *host, const char *serv)
{
	int		sockfd, n;
	struct addrinfo	hints, *res = NULL, *ressave = NULL;
	int		saved_errno = 0;

	if ((host == NULL) && (serv == NULL)) {
		print_syslog(LOG_ERR,
			"tcp_connect(): host or service are NULL.\n");
		return (-1); /* error */
	}

	memset(&hints, 0, sizeof (struct addrinfo));
#if RFC1431_WORKING_WITH_INET6
	/* Need to port rfc1431.c before using INET6: */
	hints.ai_family = AF_UNSPEC;
#else
	hints.ai_family = AF_INET;
#endif
	hints.ai_socktype = SOCK_STREAM;

	if ((n = getaddrinfo(host, serv, &hints, &res)) != 0) {
		print_syslog(LOG_ERR, "tcp_connect error for %s, %s: %s\n",
			host, serv, gai_strerror(n));
	}
	ressave = res;

	do {
		sockfd = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol);
		if (sockfd < 0) {
			saved_errno = errno;
			continue;	/* ignore this one */
		}

		if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0) {
			break;		/* success */
		}

		saved_errno = errno;
		(void) close(sockfd); /* error: close and try next one */
	} while ((res = res->ai_next) != NULL);

	if (res == NULL) {	/* errno from final socket() or connect() */
		print_syslog(LOG_ERR,
			"tcp_connect(host %s, service %s) failed;"
			" last errno %d %s",
			host? host : "(any)", serv? serv : "(any)",
			errno, strerror(errno));
		log_to_daemonlog(-1, NULL, NULL, HTTP_INTERNAL_SERVER_ERROR, 0);
		return (-1); /* error */
	}


	freeaddrinfo(ressave);

	return (sockfd);
}	/* tcp_connect() */

#else
/* XXX: Not tested */
tcp_connect(const char *host, const char *serv) {
	int			sockfd, n;
	struct sockaddr_in	servaddr;
	struct in_addr		in_val;
	struct servent		*sp = NULL;

	if ((host == NULL) && (serv == NULL)) {
		print_syslog(LOG_ERR,
			"tcp_connect(): host or service are NULL.\n");
		return (-1); /* error */
	}

	/* initialize a TCP/IP socket */
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0) == -1)) {
		print_syslog(LOG_ERR, "socket() failed; errno %d %s",
			errno, strerror(errno));
		log_to_daemonlog(-1, NULL, NULL, HTTP_INTERNAL_SERVER_ERROR, 0);
		return (-1); /* error */
	}

	memset(&servaddr, 0, sizeof (struct addrinfo));
	servaddr.sin_family = AF_INET;

	/* Lookup service or convert string port number to a number: */
	if (isdigit(serv[0])) { /* port number */
		if ((servaddr.sin_port = htons(atoi(serv))) <= 0) {
			print_syslog(LOG_ERR, "Invalid port %s", serv);
			log_to_daemonlog(-1, NULL, NULL,
					HTTP_INTERNAL_SERVER_ERROR, 0);
			return (-1); /* error */
		}
	} else { /* service name */
		if ((sp = getservbyname(serv, "tcp")) == NULL) {
			print_syslog(LOG_ERR, "Unknown service %s", serv);
			log_to_daemonlog(-1, NULL, NULL,
					HTTP_INTERNAL_SERVER_ERROR, 0);
			return (-1); /* error */
		}
		servaddr.sin_port = sp->s_port;
	}

	/* Lookup host or IP address */
	/* Use inet_aton() instead of inet_pton() for legacy code: */
#ifdef HAVE_INET_PTON
	if (inet_pton(AF_INET, host, &servaddr.sin_addr) != 0)
#else
	if (inet_aton(host, &in_val) != 0)
#endif
	{
		print_syslog(LOG_ERR, "Host %s not found", host);
		log_to_daemonlog(-1, NULL, NULL,
				HTTP_INTERNAL_SERVER_ERROR, 0);
		return (-1); /* error */
	}

	/* Connect to a host and port */
	if (connect(sockfd, (struct sockaddr *)&servaddr,
					sizeof (servaddr)) == -1) {
		print_syslog(LOG_ERR, "connect(%d) failed; errno %d %s",
			sockfd, errno, strerror(errno));
		log_to_daemonlog(-1, NULL, NULL, HTTP_INTERNAL_SERVER_ERROR, 0);
		return (-1); /* error */
	}

}	/* tcp_connect() */
#endif
#endif


#ifndef INET_ONLY
/*
 * This function does setup for servers.
 * Listen for a socket connection. That is:
 * 1.	Create a socket,
 * 2.	bind to the server's well-known port, and
 * 3.	allow * incoming requests to be accepted.
 *
 * host or service may be NULL, but not both.
 * It's very unusual to have service NULL (as that listens to EVERY port).
 *
 * See W. Richard Stevens _UNIX Network Programming_, 2d ed. v.1, p. 288.
 */
#ifdef HAVE_GETADDRINFO
int
tcp_listen(const char *host, const char *serv, socklen_t *addrlenp)
{
	int			listenfd, n;
	const int		on = 1;
	struct addrinfo		hints;
	struct addrinfo		*res = NULL;
	struct addrinfo		*ressave = NULL;
	int			saved_errno = 0;

	if ((host == NULL) && (serv == NULL)) {
		print_syslog(LOG_ERR,
			"tcp_listen(): host or service must be specified.\n");
		return (-1); /* error */
	}

	(void) memset(&hints, 0, sizeof (hints));
	hints.ai_flags = AI_PASSIVE;
#if RFC1431_WORKING_WITH_INET6
	/* Need to port rfc1431.c before using INET6: */
	hints.ai_family = AF_UNSPEC;
#else
	hints.ai_family = AF_INET;
#endif
	hints.ai_socktype = SOCK_STREAM;

	if ((n = getaddrinfo(host, serv, &hints, &res)) != 0) {
		print_syslog(LOG_ERR,
			"getaddrinfo() error for host %s, service %s: %d %s\n",
			host? host : "(any)", serv? serv : "(any)",
			n, gai_strerror(n));
		return (-1); /* error */
	}
	ressave = res;

	do {
		listenfd = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol);
		if (listenfd < 0) {
			saved_errno = errno;
			continue; /* error: try next one */
		}

		if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on,
				sizeof (on)) == -1) {
			print_syslog(LOG_ERR,
				"setsockopt(%d) failed; errno %d %s\n",
				listenfd, errno, strerror(errno));
		}

		if (bind(listenfd, res->ai_addr, res->ai_addrlen) == 0) {
			break;	/* success */
		}

		saved_errno = errno;
		(void) close(listenfd); /* bind error: close and try next one */
	} while ((res = res->ai_next) != NULL);


	if (res == NULL) {	/* errno from final socket() or bind() */
		print_syslog(LOG_ERR,
			"tcp_listen(host %s, service %s) failed;"
			" last errno %d %s",
			host? host : "(any)", serv? serv : "(any)",
			errno, strerror(errno));
		log_to_daemonlog(-1, NULL, NULL, HTTP_INTERNAL_SERVER_ERROR, 0);
		return (-1); /* error */
	}

	if (listen(listenfd, LISTENQ) == -1) {
		print_syslog(LOG_ERR, "listen(%d) failed; errno %d %s\n",
			listenfd, errno, strerror(errno));
		log_to_daemonlog(-1, NULL, NULL, HTTP_INTERNAL_SERVER_ERROR, 0);
		return (-1); /* error */
	}

	if (addrlenp != NULL) {
		/* return size of protocol address */
		*addrlenp = res->ai_addrlen;
	}

	freeaddrinfo(ressave);

	return (listenfd);
}	/* tcp_listen() */
#else
#error "tcp_listen() not implemented yet without struct addrinfo"
#endif
#endif


#ifndef INET_ONLY
/*
 * Fullfill one whois request on a socket.
 * Can be called from a POSIX pthread_create() or called directly.
 */
static void
whois_request(int connfd, struct sockaddr_in *client_addr) {
	FILE			*fd_in  = NULL;
	FILE			*fd_out = NULL;
	char			buff[MAXLINE];
	char			ident[MAXLINE] = "-";
#ifndef IDENTD_DISABLED
	struct sockaddr_in	our_server_addr;
	socklen_t		addr_len = sizeof (our_server_addr);
#endif

	/* Convert to fd to input & output file descriptors with fdopen(): */
	fd_in = fdopen(connfd, "r");
	if (fd_in == NULL) {
		print_syslog(LOG_ERR, "fdopen(\"r\") failed; errno %d %s\n",
			errno, strerror(errno));
		log_to_daemonlog(connfd, NULL, NULL,
			HTTP_INTERNAL_SERVER_ERROR, 0);
		return; /* error */
	}

	fd_out = fdopen(connfd, "w");
	if (fd_in == NULL) {
		print_syslog(LOG_ERR, "fdopen(\"w\") failed; errno %d %s\n",
			errno, strerror(errno));
		log_to_daemonlog(connfd, NULL, NULL,
			HTTP_INTERNAL_SERVER_ERROR, 0);
		return; /* error */
	}

	/* Disable buffering for the socket: */
	(void) setvbuf(fd_in, NULL, _IONBF, 0);

#ifndef IDENTD_DISABLED
	/* Query the identd server: */
	if (getsockname(fileno(fd_out), (struct sockaddr *)&our_server_addr,
			&addr_len) == -1) {
		print_syslog(LOG_ERR,
			"getsockname(out %d) failed; errno %d %s\n",
			fileno(fd_out), errno, strerror(errno));
		return;
	}

	(void) rfc1413(client_addr, &our_server_addr,
		ident, sizeof (ident));
#endif

	/* Perform the request: */
	(void) do_whois(fd_in, fd_out, ident);

	/* Cleanup: */
	(void) fflush(fd_out);
	(void) fclose(fd_in);
	(void) fclose(fd_out);
}	/* whois_request() */
#endif


#ifndef INET_ONLY
#ifdef HAVE_LIBPTHREAD
/*
 * Run whois service on a socket.
 * Can be referenced from a POSIX pthread_create() or called directly.
 */
static void *
whois_thread(void *arg) {
	struct	whois_thread_parameters	*param_ptr;
	struct	whois_thread_parameters	param;
	int				err;

	/* Get parameters */
	param_ptr = (struct whois_thread_parameters *)arg;
	if (param_ptr == NULL) {
		print_syslog(LOG_ERR,
			"NULL passed to whois_thread()\n");
		return (NULL);
	}
	param = *param_ptr;
	free(param_ptr);


	/*
	 * Detach thread, allowing memory allocated here to be freed
	 * at thread death:
	 */
	err = pthread_detach(pthread_self());
	if (err != 0) {
		print_syslog(LOG_ERR,
			"pthread_detach(pthread_self()) failed; errno %d %s\n",
			err, strerror(err));
		/* We can still continue with this error */
	}

	whois_request(param.connfd, param.client_addr);

	return (NULL);
}	/* whois_thread() */
#endif
#endif


#ifndef INET_ONLY
/*
 * Run in daemonized standalone mode
 * (i.e., without the assistance of inetd).
 * Never returns on success, return non-zero on error.
 */
int
run_standalone(void) {
	int				listenfd, connfd;
	socklen_t			addrlen = 0, len = 0;
	struct sockaddr			*client_addr = NULL;
#ifdef HAVE_LIBPTHREAD
	pthread_t			unused;
	struct whois_thread_parameters	*param = NULL;
#endif

	(void) daemonize();
	(void) initialize1();

	/* Startup message: */
	print_syslog(LOG_NOTICE, MYNAME " initialized\n");

	listenfd = tcp_listen(NULL, WHOIS_SERVICE, &addrlen);
	if (listenfd < 0) { /* Try port number instead of service name */
		listenfd = tcp_listen(NULL, WHOIS_PORT, &addrlen);
		if (listenfd < 0) {
			return (-1); /* error */
		}
	}

	/* Must call AFTER privledged whois port has been opened as root: */
	(void) initialize2();

	client_addr = (struct sockaddr *)malloc(addrlen);
	if (client_addr == NULL) {
		print_syslog(LOG_ERR, "malloc(%d) failed; errno %d %s\n",
			addrlen, errno, strerror(errno));
		return (-1); /* error */
	}

	/* Accept() new connections forever, as they arrive: */
	for (;;) {
		/* Accept a new connection */
		len = addrlen;
		connfd = accept(listenfd, client_addr, &len);
		if (connfd == -1) {
			print_syslog(LOG_ERR,
				"accept(%d) failed; errno %d %s\n",
				listenfd, errno, strerror(errno));
			log_to_daemonlog(connfd, NULL, NULL,
				HTTP_INTERNAL_SERVER_ERROR, 0);
		}

#ifdef HAVE_LIBPTHREAD
		/* Spawn a thread to handle this request: */
		param = malloc(sizeof (struct whois_thread_parameters));
		if (param == NULL) {
			print_syslog(LOG_ERR,
				"malloc(%d) failed; errno %d %s\n",
				sizeof (struct whois_thread_parameters),
				errno, strerror(errno));
			return (-1); /* error */
		}
		param->connfd = connfd;
		param->client_addr = (struct sockaddr_in *)client_addr;

		pthread_create(&unused, NULL, &whois_thread, (void *)param);
#else
		/* Single threaded: */
		whois_request(connfd, (struct sockaddr_in *)client_addr);
#endif
	}	/* for (;;) */
}	/* run_standalone() */
#endif
